!*****************************************************************************
Module Hamiltonian
!****************************************************************************

  Use TypeDef
  Use Normalisation, only : VecNormalise

  Implicit None

  complex(dbl), external :: zdotu
  
Contains

!****************************************************************************
  Subroutine EigenValues(EigSystem,HMatrix,iband)
!*****************************************************************************

    integer, intent(in) :: iband
    complex(dbl), dimension(:,:) :: HMatrix
    type(EigSystemT), pointer :: EigSystem

    Call CalculateEPsi(EigSystem,HMatrix,iband)

    ! Store the RitzValue in EigSystem
    EigSystem%Values(iband) = Dot_Product(EigSystem%States(:,iband), &
         &EigSystem%EPsi)

!    print *, 'Values:',EigSystem%Values(iband)


  end Subroutine EigenValues
!****************************************************************************
  Subroutine CalculateEPsi(EigSystem,HMatrix,iband)
!****************************************************************************

    integer, intent(in) :: iband
    complex(dbl), dimension(:,:) :: HMatrix
    type(EigSystemT), pointer :: EigSystem


    integer :: NSize,r,s
    complex(dbl) :: a,b

    a = cmplx(1.d0,0.d0)
    b = cmplx(0.d0,0.d0)

    NSize = EigSystem%HSize
    
    !    Evaluate HMatrix * EigenState, this matrix multiplication is done
    ! by a BLAS routine.
    Call zgemv('N',NSize,NSize,a,HMatrix,NSize,EigSystem%States(:,iband), &
         & 1,b,EigSystem%EPsi,1)

  end Subroutine CalculateEPsi

!****************************************************************************
  Subroutine PreCondition(EigSystem,MinVec,HMatrix,iband)
!****************************************************************************

    integer, intent(in) :: iband
    complex(dbl), dimension(:,:) :: HMatrix
    type(EigSystemT), pointer :: EigSystem
    type(MinVecT), pointer :: MinVec
    
    integer :: r,NSize
    real(dbl) :: Ekin_iter,x,K

    NSize = EigSystem%HSize

    !   There is a preconditioning matrix K for every iband, to calculate
    ! that we have to first find the quantity x, the ratio of the kinetic
    ! energy with the kinetic energy of that iteration. this is different
    ! for every band
    Ekin_iter=cmplx(0.d0,0.d0)
    do r = 1, NSize
       Ekin_iter = Ekin_iter + HMatrix(r,r) * &
            & conjg(EigSystem%States(r,iband))*EigSystem%States(r,iband)
    end do

    !   Now calculate x and K
    do r = 1, NSize
       x = HMatrix(r,r)/(Ekin_iter+1.d-30)
       
       K =    (27.d0+18.d0*x+12.d0*x**2+8.d0*x**3)/ &
            & (27.d0+18.d0*x+12.d0*x**2+8.d0*x**3+16.d0*x**4)

       ! Update the StDescVector
       MinVec%StDesc(r) = K*MinVec%StDesc(r)

    end do

!    print *, 'Precond:',MinVec%StDesc

  end Subroutine PreCondition

!***************************************************************************
  Subroutine ExactEnergyMinimise(EigSystem,MinVec,HMatrix,iband)
!**************************************************************************

    integer, intent(in) :: iband
    complex(dbl), dimension(:,:) :: HMatrix
    type(EigSystemT), pointer :: EigSystem
    type(MinVecT), pointer :: MinVec

    integer :: r,s,NSize
    real(dbl) :: E0,theta,EPhi
    complex(dbl) :: PhiHPsi,a,b
    complex(dbl), dimension(:), allocatable :: HPhi

    NSize = EigSystem%HSize

    a = cmplx(1.d0,0.d0)
    b = cmplx(0.d0,0.d0)

    E0 = cmplx(EigSystem%Values(iband))
    PhiHPsi = Dot_Product(MinVec%Phi,EigSystem%EPsi)

    allocate(HPhi(NSize))

    !    Evaluate Phi.HMatrix.Phi, by calling the BLAS routine. First
    ! calulate HMatrix.Phi
    Call zgemv('N',NSize,NSize,a,HMatrix,NSize,MinVec%Phi,1,b,HPhi,1)

    EPhi = Dot_Product(MinVec%Phi,HPhi)
    deallocate(HPhi)

    theta = 0.5d0*abs(atan((2.d0*real(PhiHPsi)/(E0 - EPhi +10.d-30))))

!    print *, 'theta:',theta

    EigSystem%States(:,iband) = EigSystem%States(:,iband)*cos(theta) + &
         & MinVec%Phi*sin(theta)

!    print *, 'States:',EigSystem%States(:,iband)

    ! Update the Ritz values
    Call EigenValues(EigSystem,HMatrix,iband)


  end Subroutine ExactEnergyMinimise

!***************************************************************************
!!$  Function Check(Eigsystem,iband)
!!$!***************************************************************************
!!$
!!$    integer, intent(in) :: iband
!!$    type(EigSystemT), pointer :: EigSystem
!!$
!!$    integer :: r,NSize
!!$    real(dbl) :: lambda,norm,Check
!!$    complex(dbl), dimension(:), allocatable :: psi,phi
!!$
!!$    NSize = EigSystem%HSize
!!$
!!$    allocate(phi(NSize))   ! Allocate space for the resedue phi
!!$
!!$    phi = EigSystem%EPsi - EigSystem%Values(iband)*Eigsystem%States(:,iband)
!!$
!!$    norm = Dot_Product(phi,phi)
!!$    Check = sqrt(norm)
!!$
!!$
!!$  end Function Check
!!$
!!$!****************************************************************************


end Module Hamiltonian

!****************************************************************************
