!*************************************************************************
Module GVectorsMod
!*************************************************************************
!
! Information related to the reciprocal lattice vectors (G-vectors).
!
! The module contains the public functions:
!
!	GVectorsInit    reads in EnergyCutOff and allocates space for gVectors,
!                   gVecIndex, and kineticE. It initializes gVectors,
!                   and gVecIndex. Finally it calls StructureFactorInit.
!
!	GVectorsDestroy deallocates space for gVectors, gVecIndex, and kineticE.
!
! The module contains the public variables:
!
!    energyCutOff         plane-wave energy cutoff (a.u.) for wavefunctions
!
!    numGVectors          total number of G-vectors  
!
!    gVectors(i,j)        j-th component (reciprocal lattice coordinates)
!                         of the i-th G-vector
!
!    gVecIndex(i,j,k)     index for the G-vectors: 
!                         gVectors(gVecIndex(i,j,k),:) is the G-vector
!                         with reciprocal lattice coordinates (i,j,k)
!
!    kineticE(i)          kineticE(gVecIndex(i,j,k)) is the kinetic energy
!                         of the G-vector with reciprocal lattice
!                         coordinates (i,j,k)
!
!    structureFactor(i,j) complex structure factor for atom type j and
!                         G-vector i
!
! The module contains the private functions:
!
!	StructureFactorInit allocates space for structureFactor and then
!                       initializes structureFactor.

  use SysParams,	only : double

  implicit none

  save

  type gVectorObjectsT
    integer                  :: numGVectors     
    integer, pointer         :: gVectors(:,:)   
    integer, pointer         :: gVecIndex(:,:,:) 
    real(double)             :: energyCutOff    
    integer, dimension(3)    :: maxGLen      
    real(double), pointer    :: kineticE(:)     
    complex(double), pointer :: structureFactor(:,:)
  end type gVectorObjectsT

  private StructureFactorInit
    
  contains

!*************************************************************************
  Subroutine GVectorsInit( gVectors, structure1, tagHandler )
!*************************************************************************
!
! GVectorInit reads in the "EnergyCutOff" and creates the G-vectors lying
! in a sphere of radius sqrt(two * maxEnergy) where maxEnergy is the
! cutoff for the potential which is 4 * EnergyCutOff for the wavefunctions.
! It creates the corresponding index gVecIndex, and 
! calls StructureFactorinit.
!
! Defined Tags:
!     EnergyCutOff

    use SysParams,	      only : four, two, twopi, electronMass, hBar2
    use TagHandlerMod,	      only : FindTag, tagHandlerT
    use StructureMod, only : StructureT
    
    implicit none

    type( StructureT ), pointer :: structure1
    type( tagHandlerT ), pointer :: tagHandler
    type( gVectorObjectsT ), pointer :: gVectors

    real(double)              :: kinetic, maxEnergy, maxGradius
    integer                   :: i, j, k, maxNumGVec, error
    integer,      allocatable :: gVectors_temp(:,:)
    integer,      dimension(3):: g_trial 
    real(double), allocatable :: kineticE_temp(:)
    character :: c

    allocate( gVectors, stat = error )
    if( error /= 0 ) stop 'Error: gVectors allocation failed.'

    !**********************************************************************
    ! Read in EnergyCutOff
    call FindTag( tagHandler,'EnergyCutOff', error)

    if(error .eq. 0) then

      read(unit = tagHandler%fileno, fmt = '(G15.10)', iostat = error) gVectors%energyCutOff
      if(error /= 0) stop 'Error, couldn''t read EnergyCutOff.'

      print *, 'EnergyCutOff in Hartrees ', gVectors%energyCutOff

    else

      stop 'Error, could not find EnergyCutOff.'

    endif
    ! energyCutOff has been initialized.
    !**********************************************************************

    !**********************************************************************
    ! Initialize gVectors, kineticE and gVecIndex.
    gVectors%numGVectors = 0
    gVectors%maxGLen     = 0
    
    ! The maximum energy for a G vector is four times the cut off energy.
    maxEnergy   = four * gVectors%energyCutOff 
    maxGradius =  sqrt(two * maxEnergy)

    ! The maximum length maxGLen(i) of a G vector along the ith reciprocal
    ! lattice vector bLatVec(i).  Think about it.
    gVectors%maxGLen(1:structure1%ndim) = int(maxGradius * structure1%lattice%aLatVecLen / twopi) + 1

    ! The number of G vectors in the parallel piped constructed with
    ! int(2*maxGLen(i)) + 1 cells in each direction.
    maxNumGVec              = product(2 * gVectors%maxGLen(1:structure1%ndim) + 1)

    allocate(gVectors%gVecIndex(-gVectors%maxGLen(1):gVectors%maxGLen(1), &
           & -gVectors%maxGLen(2):gVectors%maxGLen(2), &
           & -gVectors%maxGLen(3):gVectors%maxGLen(3)),&
           & stat=error)
    if(error /= 0) stop 'Error allocating gVectors%gVecIndex in GVectorsInit.'

    gVectors%gVecIndex   = 0

    allocate(gVectors_temp(structure1%ndim, maxNumGVec), &
           & kineticE_temp(maxNumGVec), &
           & stat=error)
    if(error /= 0) stop 'Error allocating temporary space in GVectorInit.'

    ! Loop over all cells in the parallelpiped containing the sphere of 
    ! G vectors with energy less than maxEnergy.
    do i = -gVectors%maxGLen(1), gVectors%maxGLen(1)
      do j = -gVectors%maxGLen(2), gVectors%maxGLen(2)
        do k = -gVectors%maxGLen(3), gVectors%maxGLen(3)

          g_trial = (/ i, j, k /)
          kinetic = dot_product(g_trial(1:structure1%ndim), &
                   & matmul(structure1%lattice%bLatMetricTensor,g_trial(1:structure1%ndim)))/two 

          ! If the energy is less than maxEnergy store the G vector for
          ! later StructureMod.
          if(kinetic < maxEnergy) then

            gVectors%numGVectors = gVectors%numGVectors + 1
            gVectors_temp(:,gVectors%numGVectors) = g_trial(1:structure1%ndim)
            kineticE_temp(gVectors%numGVectors) = kinetic
            gVectors%gVecIndex(i,j,k) = gVectors%numGVectors  ! determine index array  

          end if
        end do
      end do
    end do

    ! allocate final gVectors array and copy over from temporary storage
    allocate(gVectors%gVectors(structure1%ndim, gVectors%numGVectors), &
           & gVectors%kineticE(gVectors%numGVectors), stat=error)
    if(error /= 0) stop 'Error allocating space for G vectors in GVectorsInit.'

    ! transfer vectors to permanent storage
    gVectors%gVectors(:,1:gVectors%numGVectors) = gVectors_temp(:,1:gVectors%numGVectors) 
    gVectors%kineticE(1:gVectors%numGVectors)   = kineticE_temp(1:gVectors%numGVectors)

    deallocate(gVectors_temp,kineticE_temp) ! deallocate associated work space
    ! gVectors, kineticE, and gVecIndex initialized.
    !**********************************************************************

    ! Initialize the structure factors.
    call StructureFactorInit( gVectors, structure1 )

  end Subroutine GVectorsInit

!*************************************************************************
  Subroutine GVectorsDestroy( gVectors )
!*************************************************************************
!
! Deallocates the space allocated for the GVectors.

    implicit none

    type( gVectorObjectsT ), pointer :: gVectors
    integer error

    if( .not. associated( gVectors )) stop 'Error: gVectors not allocated.'

    deallocate(gVectors%gVectors, gVectors%kineticE, stat = error)
    if(error /= 0) stop 'Error gVectors deallocating in GVectorsDestroy.'

    deallocate(gVectors%gVecIndex, stat = error)
    if(error /= 0) stop 'Error gVectors%gVecIndex deallocating in GVectorsDestroy.'

    deallocate(gVectors, stat = error)
    if(error /= 0) stop 'Error gVectors deallocating in GVectorsDestroy.'

  end Subroutine GVectorsDestroy

!*************************************************************************
  Subroutine StructureFactorInit( gVectors, structure1 )
!*************************************************************************
!
!   calculates the structure factors for each atom type at each G-vector

    use SysParams,        only : one, twopi, double
    use StructureMod, only : StructureT

    implicit none

    type( gVectorObjectsT ), pointer :: gVectors
    type( StructureT ), pointer :: structure1

    real(double) :: arg 
    integer      :: iGVec, iElem , iAtom, error, atomI

    allocate (gVectors%structureFactor(gVectors%numGVectors, structure1%atomBasis%numSpecies),stat=error)
    if(error /= 0) stop 'structureFactor allocation error'

    atomI = 0
    gVectors%structureFactor = 0

    ! Structure Factor exist for each G vector and each element.  Each
    ! structure factor has a contribution from each atom of that element =
    ! exp(i* 2 * pi (G * R))
    do iElem = 1, structure1%atomBasis%numSpecies

      do iAtom = 1, structure1%atomBasis%numAtomsOfSpecies(iElem)
    
        atomI = atomI + 1
    
        do iGVec = 1, gVectors%numGVectors                             

          arg = twopi * dot_product(gVectors%gVectors(:,iGVec), structure1%atomBasis%atomLat(:,atomI))
          gVectors%structureFactor(iGVec,iElem) = gVectors%structureFactor(iGVec,iElem) + &
                                   cmplx(cos(arg),-sin(arg),kind=double)

        end do
      end do
    end do

    ! Normalize the structure factors.
    gVectors%structureFactor = gVectors%structureFactor / structure1%atomBasis%totalNumAtoms  

  end Subroutine StructureFactorInit

!*************************************************************************
  Subroutine GVectorScatter( gVectors, structure1, eigenStates, densityArray, k, n )
!*************************************************************************

    use StructureMod,   only : StructureT
    use EigenStatesMod, only : eigenStatesT
	use densityArrayMod, only : densityArrayT

    implicit none

    type( gVectorObjectsT ),      pointer :: gVectors
    type( StructureT ),     pointer :: structure1
    type( eigenStatesT ),   pointer :: eigenStates
    type( densityArrayT ), pointer :: densityArray

    integer, intent(in) :: k, n
    integer             :: i, j, index

    densityArray%waveFunction = 0

    do i = 1, eigenStates%numBasisVectors(k)  

      if(gVectors%gVectors(structure1%ndim, eigenStates%basisIndex(i,k)) .ge. 0) then
        index = gVectors%gVectors(structure1%ndim, eigenStates%basisIndex(i,k))
      else
        index = gVectors%gVectors(structure1%ndim, eigenStates%basisIndex(i,k)) + &
              & densityArray%gridDimensions(structure1%ndim)
      end if

      do j = structure1%ndim-1, 1, -1
        if(gVectors%gVectors(j, eigenStates%basisIndex(i,k)) .ge. 0) then
          index = index * densityArray%gridDimensions(j+1) + gVectors%gVectors(j, eigenStates%basisIndex(i,k))
        else
          index = index * densityArray%gridDimensions(j+1) + gVectors%gVectors(j, eigenStates%basisIndex(i,k)) &
                & + densityArray%gridDimensions(j)
        end if
      end do  
      index = index + 1
      densityArray%waveFunction(index) = eigenStates%eigenVectors(i, n, k)

    end do

  end Subroutine GVectorScatter

end Module GVectorsMod
