!*************************************************************************
Program PW
!*************************************************************************
! PW calculates the band structure along various directions in
! k-space using a plane-wave basis and a fixed "effective" potential.  
! Several choices of empirical pseudopotentials are provided.
! Written by William Mattson and Richard M. Martin of the University 
! of Illinois, based upon program written by K. Glassford and I. Souza.

! Note: atomic units (a.u.) are used throughout the program.

  use TagHandlerMod!,       only : tagHandlerT, TagHandlerInit, TagHandlerDestroy 
  use StructureMod,        only : StructureT, StructureInit, StructureDestroy
  use kPointsMod,          only : kPointsT, KPointsInit, KPointsDestroy
  use graphMod,            only : graphT, PlotBands
  use HamiltonianMod,      only : hamiltonianArrayT, HInit, HDestroy, HDiagonalize,  & 
                                  & HPrint, ParamsT
  use EigenStatesMod,      only : eigenStatesT, EigenStatesInit, EigenStatesDestroy, &
                                  & PrintEigenvalues
  use SpecificHamMod,      only : hamInfoT, HamInfoInit, HamInfoDestroy, HGenerate, FindHMaxSize
!  use atomBasisMod,        only : AtomsPrint  

  implicit none

  type(hamInfoT),          pointer :: hamInfo
  type(StructureT),        pointer :: structure1
  type(kPointsT),          pointer :: kPoints
  type(eigenStatesT),      pointer :: eigenStates
  type(hamiltonianArrayT), pointer :: hamiltonian
  type(tagHandlerT),       pointer :: tagHandler
  type(graphT),            pointer :: graphinfo
  type(ParamsT)                    :: Params

  integer :: kpt, r
  integer :: hamArraysize
  character(80) :: filename

  print *, ' Input file name' 
  read( 5, '(A80)') filename

  call TagHandlerInit( tagHandler, 10, filename) ! Initialize "tag handler" to read
                                                   ! the input from file filename
  Call ParamsInit(Params,TagHandler)

  call StructureInit( structure1, tagHandler )       ! calls DimensionsInit, LatticeInit, AtomsInit
!  call AtomsPrint( structure1%atomBasis )

  call KPointsInit( kPoints, graphInfo, structure1%ndim, structure1%lattice, &
       & tagHandler ,Params%conv, Params%period)

  !  If we are using Conjugate Gradient then open the file for the 
  ! convergence data
  if(Params%conv.eq.1) then
     open(unit=9,file="convergence.dat",status="replace",position="rewind")
     close(9)
  end if

  call HamInfoInit( hamInfo, structure1, tagHandler )  ! Gvector list and structurefactor

  call FindHMaxSize( hamInfo, structure1, kPoints )	   ! Find hMaxSize = max size of H at any k point
  
  hamArraysize = hamInfo%hMaxSize 

  call HInit( hamiltonian, hamArraysize )		! Allocates array for H and work arrays

  call EigenStatesInit( eigenStates, kPoints%numBands, kPoints%numKPoints, hamArraysize)

  eigenStates%eigenVectors = 0

  if(Params%Switch.eq.1) then
     open(unit=110,file='Report.cg',status='replace',action="write", & 
          & position="rewind")
     write(110,*) 'Eigenvalues:'
     close(unit=110)
  end if

  do kpt = 1, kPoints%numKPoints                             ! loop over k-vectors

     call HGenerate( hamInfo, structure1, kPoints, eigenStates, &
          & kpt, hamiltonian%h, hamiltonian%SqH,hamiltonian%hSize )

     if(Params%Switch.eq.1) then
        open(unit=110,file='Report.cg',status='old',action="write", & 
             & position="append")
        write(110,*) ''
        write(110,*) 'k:',kpt
        close(unit=110)
     end if

      call HDiagonalize( hamiltonian, eigenStates, kpt , Params )

      print *, kpt,'/',kPoints%numKPoints,'completed'

    end do  

! Creates files plot.dat and gnuplot.dat (Display with: gnuplot gnuplot.dat)

 if( associated( graphinfo ) )    call PlotBands( graphinfo, eigenStates )           

 if( .not. associated( graphinfo ) )  call PrintEigenvalues(eigenStates, kPoints%numKPoints)    

  call EigenStatesDestroy( eigenStates )		!  Clean up memory

  call HDestroy( hamiltonian )
                
  call HamInfoDestroy( hamInfo )

  call KPointsDestroy( kPoints )

  call StructureDestroy( structure1 )

  call TagHandlerDestroy( tagHandler )     ! Clean up and close the input file

Contains

!************************************************************************************
  Subroutine ParamsInit(Params,TagHandler)
!************************************************************************************

    type(TagHandlerT), pointer :: TagHandler
    type(ParamsT) :: Params    

    integer :: error,error1
    logical :: default
    
    ! Fille in the tolerance in Params%tol
    default = .True.
    Call FindTag(TagHandler,'CGTolerance',error1)
    if(error1.eq.0) then
       read(unit=tagHandler%fileno, fmt = '(G15.10)', iostat=error ) & 
            & Params%tol
       if(error.eq.0 .and. Params%tol.ne.0 ) default = .False.
    end if
    if(default) then
       Params%tol = 1.d-5
       print *, 'Using the default value of CGTolerance',Params%tol
    end if

    ! Fille in the tolerance in Params%switch
    default = .True.
    Call FindTag(TagHandler,'DiagonalizationSwitch',error1)
    if(error1.eq.0) then
       read(unit=tagHandler%fileno, fmt = '(I10)', iostat=error ) & 
            & Params%Switch
       if(error.eq.0 .and. Params%Switch.eq.0 .or. Params%Switch.eq.1 ) &
            & default = .False.
    end if
    if(default) then
       Params%Switch = 0
       print *, 'Using the Lapack routine for Diagonalisation'
    end if

    if(Params%Switch.eq.1) then

       ! Fille in the tolerance in Params%period
       default = .True.
       Call FindTag(TagHandler,'CGIterationPeriod',error1)
       if(error1.eq.0) then
          read(unit=tagHandler%fileno, fmt = '(I10)', iostat=error ) & 
               & Params%period
          if(error.eq.0 .and. Params%period.gt.0 ) default = .False.
       end if
       if(default) then
          Params%period = 3
          print *, 'Using the default value of CGIterationPeriod',Params%period
       end if

       ! Fille in the preconditioning in Params%precond
       default = .True.
       Call FindTag(TagHandler,'CGPreConditioner',error1)
       if(error1.eq.0) then
          read(unit=tagHandler%fileno, fmt = '(I10)', iostat=error ) & 
               & Params%precond
          if(error.eq.0 .and. (Params%precond.eq.0 .or. Params%precond.eq.1)) default = .False.
       end if
       if(default) then
          Params%precond = 0
          print *, 'Using the default Preconditioner',Params%precond
       end if

       ! Fille in the convergence switch in Params%conv
       default = .True.
       Call FindTag(TagHandler,'CGConvergenceSwitch',error1)
       if(error1.eq.0) then
          read(unit=tagHandler%fileno, fmt = '(I10)', iostat=error ) & 
               & Params%conv
          if(error.eq.0 .and. (Params%conv.eq.0 .or. Params%conv.eq.1)) default = .False.
       end if
       if(default) then
          Params%conv = 0
          print *, 'Not testing for convergence'
       end if

    end if

  end Subroutine ParamsInit

!************************************************************************************


end Program PW
